@@ -26,7 +26,6 @@
 #include <stdlib.h>
 #include <limits.h>
 #include <poll.h>
-#include <siphash.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
@@ -64,6 +63,7 @@ static struct repo_read {
 	int *temp_fds;
 	int session_fd;
 	struct gotd_imsgev session_iev;
+	int refs_listed;
 } repo_read;
 
 static struct repo_read_client {
@@ -141,7 +141,6 @@ send_symref(struct got_reference *symref, struct got_object_id *target_id,
 		goto done;
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 done:
 	free(target_id);
@@ -198,7 +197,6 @@ send_peeled_tag_ref(struct got_reference *ref, struct got_object *obj,
 		goto done;
 	}
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 done:
 	got_object_tag_close(tag);
@@ -241,7 +239,6 @@ send_ref(struct got_reference *ref, struct imsgbuf *ibuf)
 	if (imsg_add(wbuf, refname, namelen) == -1)
 		return got_error_from_errno("imsg_add REF");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 
 	err = got_object_open(&obj, repo_read.repo, id);
@@ -263,30 +260,28 @@ list_refs(struct imsg *imsg)
 	struct repo_read_client *client = &repo_read_client;
 	struct got_reflist_head refs;
 	struct got_reflist_entry *re;
-	struct gotd_imsg_list_refs_internal ireq;
 	size_t datalen;
 	struct gotd_imsg_reflist irefs;
 	struct imsgbuf ibuf;
-	int client_fd = imsg->fd;
+	int client_fd;
 	struct got_object_id *head_target_id = NULL;
 
 	TAILQ_INIT(&refs);
 
+	client_fd = imsg_get_fd(imsg);
 	if (client_fd == -1)
 		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
 	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
-	if (datalen != sizeof(ireq))
+	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
-	memcpy(&ireq, imsg->data, sizeof(ireq));
 
-	if (ireq.client_id == 0)
-		return got_error(GOT_ERR_CLIENT_ID);
-	if (client->id != 0) {
+	if (repo_read.refs_listed) {
 		return got_error_msg(GOT_ERR_CLIENT_ID,
 		    "duplicate list-refs request");
 	}
-	client->id = ireq.client_id;
+	repo_read.refs_listed = 1;
+
 	client->fd = client_fd;
 
 	imsg_init(&ibuf, client_fd);
@@ -562,9 +557,6 @@ receive_delta_cache_fd(struct imsg *imsg,
 
 	log_debug("receiving delta cache file");
 
-	if (imsg->fd == -1)
-		return got_error(GOT_ERR_PRIVSEP_NO_FD);
-
 	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
 	if (datalen != sizeof(ireq))
 		return got_error(GOT_ERR_PRIVSEP_LEN);
@@ -573,7 +565,10 @@ receive_delta_cache_fd(struct imsg *imsg,
 	if (client->delta_cache_fd != -1)
 		return got_error(GOT_ERR_PRIVSEP_MSG);
 
-	client->delta_cache_fd = imsg->fd;
+	client->delta_cache_fd = imsg_get_fd(imsg);
+	if (client->delta_cache_fd == -1)
+		return got_error(GOT_ERR_PRIVSEP_NO_FD);
+
 	client->report_progress = ireq.report_progress;
 	return NULL;
 }
@@ -582,23 +577,21 @@ static const struct got_error *
 receive_pack_pipe(struct imsg *imsg, struct gotd_imsgev *iev)
 {
 	struct repo_read_client *client = &repo_read_client;
-	struct gotd_imsg_packfile_pipe ireq;
 	size_t datalen;
 
 	log_debug("receiving pack pipe descriptor");
 
-	if (imsg->fd == -1)
-		return got_error(GOT_ERR_PRIVSEP_NO_FD);
-
 	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
-	if (datalen != sizeof(ireq))
+	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
-	memcpy(&ireq, imsg->data, sizeof(ireq));
 
 	if (client->pack_pipe != -1)
 		return got_error(GOT_ERR_PRIVSEP_MSG);
 
-	client->pack_pipe = imsg->fd;
+	client->pack_pipe = imsg_get_fd(imsg);
+	if (client->pack_pipe == -1)
+		return got_error(GOT_ERR_PRIVSEP_NO_FD);
+
 	return NULL;
 }
 
@@ -607,7 +600,6 @@ send_packfile(struct imsg *imsg, struct gotd_imsgev *iev)
 {
 	const struct got_error *err = NULL;
 	struct repo_read_client *client = &repo_read_client;
-	struct gotd_imsg_packfile_done idone;
 	uint8_t packsha1[SHA1_DIGEST_LENGTH];
 	char hex[SHA1_DIGEST_STRING_LENGTH];
 	FILE *delta_cache = NULL;
@@ -660,12 +652,14 @@ send_packfile(struct imsg *imsg, struct gotd_imsgev *iev)
 	    got_sha1_digest_to_str(packsha1, hex, sizeof(hex)))
 		log_debug("sent pack-%s.pack", hex);
 
-	memset(&idone, 0, sizeof(idone));
-	idone.client_id = client->id;
 	if (gotd_imsg_compose_event(iev, GOTD_IMSG_PACKFILE_DONE,
-	    PROC_REPO_READ, -1, &idone, sizeof(idone)) == -1)
+	    PROC_REPO_READ, -1, NULL, 0) == -1)
 		err = got_error_from_errno("imsg compose PACKFILE_DONE");
 done:
+	if (client->delta_cache_fd != -1 &&
+	    close(client->delta_cache_fd) == -1 && err == NULL)
+		err = got_error_from_errno("close");
+	client->delta_cache_fd = -1;
 	if (delta_cache != NULL && fclose(delta_cache) == EOF && err == NULL)
 		err = got_error_from_errno("fclose");
 	imsg_clear(&ibuf);
@@ -707,7 +701,7 @@ repo_read_dispatch_session(int fd, short event, void *arg)
 			break;
 
 		if (imsg.hdr.type != GOTD_IMSG_LIST_REFS_INTERNAL &&
-		    client->id == 0) {
+		    !repo_read.refs_listed) {
 			err = got_error(GOT_ERR_PRIVSEP_MSG);
 			break;
 		}
@@ -776,13 +770,13 @@ recv_connect(struct imsg *imsg)
 	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
 	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
-	if (imsg->fd == -1)
-		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
 	if (repo_read.session_fd != -1)
 		return got_error(GOT_ERR_PRIVSEP_MSG);
 
-	repo_read.session_fd = imsg->fd;
+	repo_read.session_fd = imsg_get_fd(imsg);
+	if (repo_read.session_fd == -1)
+		return got_error(GOT_ERR_PRIVSEP_NO_FD);
 
 	imsg_init(&iev->ibuf, repo_read.session_fd);
 	iev->handler = repo_read_dispatch_session;
@@ -923,7 +917,7 @@ repo_read_shutdown(void)
 {
 	struct repo_read_client *client = &repo_read_client;
 
-	log_debug("shutting down");
+	log_debug("%s: shutting down", repo_read.title);
 
 	if (client->have_ids)
 		got_object_idset_free(client->have_ids);
