@@ -40,6 +40,7 @@
 #include "got_privsep.h"
 
 #include "gotwebd.h"
+#include "log.h"
 
 static const struct got_error *got_init_repo_commit(struct repo_commit **);
 static const struct got_error *got_init_repo_tag(struct repo_tag **);
@@ -383,7 +384,7 @@ got_get_repo_commits(struct request *c, size_t limit)
 			goto done;
 	}
 
-	error = got_commit_graph_iter_start(graph, id, repo, NULL, NULL);
+	error = got_commit_graph_bfsort(graph, id, repo, NULL, NULL);
 	if (error)
 		goto done;
 
@@ -944,8 +945,7 @@ got_gotweb_blame_cb(void *arg, int nlines, int lineno,
 	committer_time = got_object_commit_get_committer_time(commit);
 	if (gmtime_r(&committer_time, &tm) == NULL)
 		return got_error_from_errno("gmtime_r");
-	if (strftime(bline->datebuf, sizeof(bline->datebuf), "%G-%m-%d",
-	    &tm) == 0) {
+	if (strftime(bline->datebuf, sizeof(bline->datebuf), "%F", &tm) == 0) {
 		err = got_error(GOT_ERR_NO_SPACE);
 		goto done;
 	}
@@ -1005,8 +1005,7 @@ got_output_file_blame(struct request *c, got_render_blame_line_cb cb)
 	bca.lines = NULL;
 	bca.cb = cb;
 
-	if (asprintf(&path, "%s%s%s", qs->folder ? qs->folder : "",
-	    qs->folder ? "/" : "", qs->file) == -1) {
+	if (asprintf(&path, "%s/%s", qs->folder, qs->file) == -1) {
 		error = got_error_from_errno("asprintf");
 		goto done;
 	}
