@@ -49,6 +49,7 @@
 #include "got_privsep.h"
 
 #include "gotwebd.h"
+#include "log.h"
 #include "tmpl.h"
 
 static const struct querystring_keys querystring_keys[] = {
@@ -59,7 +60,6 @@ static const struct querystring_keys querystring_keys[] = {
 	{ "headref",		HEADREF },
 	{ "index_page",		INDEX_PAGE },
 	{ "path",		PATH },
-	{ "page",		PAGE },
 };
 
 static const struct action_keys action_keys[] = {
@@ -80,15 +80,13 @@ static const struct action_keys action_keys[] = {
 };
 
 static const struct got_error *gotweb_init_querystring(struct querystring **);
-static const struct got_error *gotweb_parse_querystring(struct querystring **,
+static const struct got_error *gotweb_parse_querystring(struct querystring *,
     char *);
-static const struct got_error *gotweb_assign_querystring(struct querystring **,
+static const struct got_error *gotweb_assign_querystring(struct querystring *,
     char *, char *);
 static int gotweb_render_index(struct template *);
-static const struct got_error *gotweb_init_repo_dir(struct repo_dir **,
-    const char *);
-static const struct got_error *gotweb_load_got_path(struct request *c,
-    struct repo_dir *);
+static const struct got_error *gotweb_load_got_path(struct repo_dir **,
+    const char *, struct request *);
 static const struct got_error *gotweb_get_repo_description(char **,
     struct server *, const char *, int);
 static const struct got_error *gotweb_get_clone_url(char **, struct server *,
@@ -157,9 +155,6 @@ gotweb_process_request(struct request *c)
 		log_warnx("%s: %s", __func__, error->msg);
 		return;
 	}
-	/* don't process any further if client disconnected */
-	if (c->sock->client_status == CLIENT_DISCONNECT)
-		return;
 	/* get the gotwebd server */
 	srv = gotweb_get_server(c->server_name);
 	if (srv == NULL) {
@@ -174,7 +169,7 @@ gotweb_process_request(struct request *c)
 		goto err;
 	}
 	c->t->qs = qs;
-	error = gotweb_parse_querystring(&qs, c->querystring);
+	error = gotweb_parse_querystring(qs, c->querystring);
 	if (error) {
 		log_warnx("%s: %s", __func__, error->msg);
 		goto err;
@@ -196,16 +191,18 @@ gotweb_process_request(struct request *c)
 	}
 
 	if (qs->action != INDEX) {
-		error = gotweb_init_repo_dir(&repo_dir, qs->path);
-		if (error)
-			goto err;
-		error = gotweb_load_got_path(c, repo_dir);
+		error = gotweb_load_got_path(&repo_dir, qs->path, c);
 		c->t->repo_dir = repo_dir;
 		if (error && error->code != GOT_ERR_LONELY_PACKIDX)
 			goto err;
 	}
 
 	if (qs->action == BLOBRAW || qs->action == BLOB) {
+		if (qs->folder == NULL || qs->file == NULL) {
+			error = got_error(GOT_ERR_BAD_QUERYSTRING);
+			goto err;
+		}
+
 		error = got_get_repo_commits(c, 1);
 		if (error)
 			goto err;
@@ -218,6 +215,10 @@ gotweb_process_request(struct request *c)
 
 	switch (qs->action) {
 	case BLAME:
+		if (qs->folder == NULL || qs->file == NULL) {
+			error = got_error(GOT_ERR_BAD_QUERYSTRING);
+			goto err;
+		}
 		error = got_get_repo_commits(c, 1);
 		if (error) {
 			log_warnx("%s: %s", __func__, error->msg);
@@ -231,7 +232,6 @@ gotweb_process_request(struct request *c)
 		if (binary) {
 			struct gotweb_url url = {
 				.index_page = -1,
-				.page = -1,
 				.action = BLOBRAW,
 				.path = qs->path,
 				.commit = qs->commit,
@@ -347,11 +347,11 @@ gotweb_process_request(struct request *c)
 			    error->msg);
 			goto err;
 		}
-		error = got_get_repo_commits(c, D_MAXSLCOMMDISP);
+		error = got_get_repo_commits(c, srv->summary_commits_display);
 		if (error)
 			goto err;
 		qs->action = TAGS;
-		error = got_get_repo_tags(c, D_MAXSLTAGDISP);
+		error = got_get_repo_tags(c, srv->summary_tags_display);
 		if (error) {
 			log_warnx("%s: got_get_repo_tags: %s", __func__,
 			    error->msg);
@@ -459,17 +459,12 @@ gotweb_init_querystring(struct querystring **qs)
 	}
 
 	(*qs)->action = INDEX;
-	(*qs)->commit = NULL;
-	(*qs)->file = NULL;
-	(*qs)->folder = NULL;
-	(*qs)->index_page = 0;
-	(*qs)->path = NULL;
 
 	return error;
 }
 
 static const struct got_error *
-gotweb_parse_querystring(struct querystring **qs, char *qst)
+gotweb_parse_querystring(struct querystring *qs, char *qst)
 {
 	const struct got_error *error = NULL;
 	char *tok1 = NULL, *tok1_pair = NULL, *tok1_end = NULL;
@@ -564,7 +559,7 @@ gotweb_urldecode(char *url)
 }
 
 static const struct got_error *
-gotweb_assign_querystring(struct querystring **qs, char *key, char *value)
+gotweb_assign_querystring(struct querystring *qs, char *key, char *value)
 {
 	const struct got_error *error = NULL;
 	const char *errstr;
@@ -580,47 +575,43 @@ gotweb_assign_querystring(struct querystring **qs, char *key, char *value)
 
 		switch (querystring_keys[el_cnt].element) {
 		case ACTION:
-			for (a_cnt = 0; a_cnt < ACTIONS__MAX; a_cnt++) {
+			for (a_cnt = 0; a_cnt < nitems(action_keys); a_cnt++) {
 				if (strcmp(value, action_keys[a_cnt].name) != 0)
 					continue;
-				else if (strcmp(value,
-				    action_keys[a_cnt].name) == 0){
-					(*qs)->action =
-					    action_keys[a_cnt].action;
-					goto qa_found;
-				}
+				qs->action = action_keys[a_cnt].action;
+				goto qa_found;
 			}
-			(*qs)->action = ERR;
+			qs->action = ERR;
 qa_found:
 			break;
 		case COMMIT:
-			(*qs)->commit = strdup(value);
-			if ((*qs)->commit == NULL) {
+			qs->commit = strdup(value);
+			if (qs->commit == NULL) {
 				error = got_error_from_errno2(__func__,
 				    "strdup");
 				goto done;
 			}
 			break;
 		case RFILE:
-			(*qs)->file = strdup(value);
-			if ((*qs)->file == NULL) {
+			qs->file = strdup(value);
+			if (qs->file == NULL) {
 				error = got_error_from_errno2(__func__,
 				    "strdup");
 				goto done;
 			}
 			break;
 		case FOLDER:
-			(*qs)->folder = strdup(value);
-			if ((*qs)->folder == NULL) {
+			qs->folder = strdup(value);
+			if (qs->folder == NULL) {
 				error = got_error_from_errno2(__func__,
 				    "strdup");
 				goto done;
 			}
 			break;
 		case HEADREF:
-			free((*qs)->headref);
-			(*qs)->headref = strdup(value);
-			if ((*qs)->headref == NULL) {
+			free(qs->headref);
+			qs->headref = strdup(value);
+			if (qs->headref == NULL) {
 				error = got_error_from_errno2(__func__,
 				    "strdup");
 				goto done;
@@ -629,37 +620,24 @@ qa_found:
 		case INDEX_PAGE:
 			if (*value == '\0')
 				break;
-			(*qs)->index_page = strtonum(value, INT64_MIN,
+			qs->index_page = strtonum(value, INT64_MIN,
 			    INT64_MAX, &errstr);
 			if (errstr) {
 				error = got_error_from_errno3(__func__,
 				    "strtonum", errstr);
 				goto done;
 			}
-			if ((*qs)->index_page < 0)
-				(*qs)->index_page = 0;
+			if (qs->index_page < 0)
+				qs->index_page = 0;
 			break;
 		case PATH:
-			(*qs)->path = strdup(value);
-			if ((*qs)->path == NULL) {
+			qs->path = strdup(value);
+			if (qs->path == NULL) {
 				error = got_error_from_errno2(__func__,
 				    "strdup");
 				goto done;
 			}
 			break;
-		case PAGE:
-			if (*value == '\0')
-				break;
-			(*qs)->page = strtonum(value, INT64_MIN,
-			    INT64_MAX, &errstr);
-			if (errstr) {
-				error = got_error_from_errno3(__func__,
-				    "strtonum", errstr);
-				goto done;
-			}
-			if ((*qs)->page < 0)
-				(*qs)->page = 0;
-			break;
 		}
 
 		/* entry found */
@@ -780,7 +758,6 @@ gotweb_index_navs(struct request *c, struct gotweb_url *prev, int *have_prev,
 		*prev = (struct gotweb_url){
 			.action = -1,
 			.index_page = qs->index_page - 1,
-			.page = -1,
 		};
 	}
 	if (t->next_disp == srv->max_repos_display &&
@@ -790,7 +767,6 @@ gotweb_index_navs(struct request *c, struct gotweb_url *prev, int *have_prev,
 		*next = (struct gotweb_url){
 			.action = -1,
 			.index_page = qs->index_page + 1,
-			.page = -1,
 		};
 	}
 }
@@ -813,9 +789,6 @@ gotweb_render_index(struct template *tp)
 		return -1;
 
 	for (d_i = 0; d_i < t->nrepos; d_i++) {
-		if (srv->max_repos > 0 && t->prev_disp == srv->max_repos)
-			break;
-
 		if (strcmp(sd_dent[d_i]->d_name, ".") == 0 ||
 		    strcmp(sd_dent[d_i]->d_name, "..") == 0) {
 			d_skipped++;
@@ -837,11 +810,8 @@ gotweb_render_index(struct template *tp)
 			continue;
 		}
 
-		error = gotweb_init_repo_dir(&repo_dir, sd_dent[d_i]->d_name);
-		if (error)
-			continue;
-
-		error = gotweb_load_got_path(c, repo_dir);
+		error = gotweb_load_got_path(&repo_dir, sd_dent[d_i]->d_name,
+		    c);
 		if (error && error->code != GOT_ERR_LONELY_PACKIDX) {
 			if (error->code != GOT_ERR_NOT_GIT_REPO)
 				log_warnx("%s: %s: %s", __func__,
@@ -869,11 +839,7 @@ gotweb_render_index(struct template *tp)
 	}
 	t->repos_total = t->nrepos - d_skipped;
 
-	if (srv->max_repos_display == 0)
-		return 0;
-	if (srv->max_repos > 0 && srv->max_repos < srv->max_repos_display)
-		return 0;
-	if (t->repos_total <= srv->max_repos ||
+	if (srv->max_repos_display == 0 ||
 	    t->repos_total <= srv->max_repos_display)
 		return 0;
 
@@ -1010,18 +976,6 @@ gotweb_render_url(struct request *c, struct gotweb_url *url)
 		sep = "&";
 	}
 
-	if (url->previd) {
-		if (tp_writef(c->tp, "%sprevid=%s", sep, url->previd) == -1)
-			return -1;
-		sep = "&";
-	}
-
-	if (url->prevset) {
-		if (tp_writef(c->tp, "%sprevset=%s", sep, url->prevset) == -1)
-			return -1;
-		sep = "&";
-	}
-
 	if (url->file) {
 		tmp = gotweb_urlencode(url->file);
 		if (tmp == NULL)
@@ -1073,12 +1027,6 @@ gotweb_render_url(struct request *c, struct gotweb_url *url)
 		sep = "&";
 	}
 
-	if (url->page != -1) {
-		if (tp_writef(c->tp, "%spage=%d", sep, url->page) == -1)
-			return -1;
-		sep = "&";
-	}
-
 	return 0;
 }
 
@@ -1098,43 +1046,62 @@ gotweb_render_absolute_url(struct request *c, struct gotweb_url *url)
 }
 
 static const struct got_error *
-gotweb_load_got_path(struct request *c, struct repo_dir *repo_dir)
+gotweb_load_got_path(struct repo_dir **rp, const char *dir,
+    struct request *c)
 {
 	const struct got_error *error = NULL;
-	struct socket *sock = c->sock;
 	struct server *srv = c->srv;
 	struct transport *t = c->t;
+	struct repo_dir *repo_dir;
 	DIR *dt;
 	char *dir_test;
 
-	if (asprintf(&dir_test, "%s/%s/%s", srv->repos_path, repo_dir->name,
+	*rp = calloc(1, sizeof(**rp));
+	if (*rp == NULL)
+		return got_error_from_errno("calloc");
+	repo_dir = *rp;
+
+	if (asprintf(&dir_test, "%s/%s/%s", srv->repos_path, dir,
 	    GOTWEB_GIT_DIR) == -1)
 		return got_error_from_errno("asprintf");
 
 	dt = opendir(dir_test);
 	if (dt == NULL) {
 		free(dir_test);
-		if (asprintf(&dir_test, "%s/%s", srv->repos_path,
-		    repo_dir->name) == -1)
+		if (asprintf(&dir_test, "%s/%s", srv->repos_path, dir) == -1)
 			return got_error_from_errno("asprintf");
 		dt = opendir(dir_test);
 		if (dt == NULL) {
 			free(dir_test);
-			return got_error_path(repo_dir->name,
-			    GOT_ERR_NOT_GIT_REPO);
+			if (asprintf(&dir_test, "%s/%s%s", srv->repos_path,
+			    dir, GOTWEB_GIT_DIR) == -1)
+				return got_error_from_errno("asprintf");
+			dt = opendir(dir_test);
+			if (dt == NULL) {
+				free(dir_test);
+				return got_error_path(dir,
+				    GOT_ERR_NOT_GIT_REPO);
+			}
 		}
 	}
 
 	repo_dir->path = dir_test;
 	dir_test = NULL;
 
+	repo_dir->name = strdup(repo_dir->path + strlen(srv->repos_path) + 1);
+	if (repo_dir->name == NULL) {
+		error = got_error_from_errno("strdup");
+		goto err;
+	}
+
 	if (srv->respect_exportok &&
 	    faccessat(dirfd(dt), "git-daemon-export-ok", F_OK, 0) == -1) {
 		error = got_error_path(repo_dir->name, GOT_ERR_NOT_GIT_REPO);
 		goto err;
 	}
 
-	error = got_repo_open(&t->repo, repo_dir->path, NULL, sock->pack_fds);
+	error = got_repo_open(&t->repo, repo_dir->path, NULL,
+	    gotwebd_env->pack_fds);
 	if (error)
 		goto err;
 	error = gotweb_get_repo_description(&repo_dir->description, srv,
@@ -1162,29 +1129,6 @@ err:
 	return error;
 }
 
-static const struct got_error *
-gotweb_init_repo_dir(struct repo_dir **repo_dir, const char *dir)
-{
-	const struct got_error *error;
-
-	*repo_dir = calloc(1, sizeof(**repo_dir));
-	if (*repo_dir == NULL)
-		return got_error_from_errno("calloc");
-
-	if (asprintf(&(*repo_dir)->name, "%s", dir) == -1) {
-		error = got_error_from_errno("asprintf");
-		free(*repo_dir);
-		*repo_dir = NULL;
-		return error;
-	}
-	(*repo_dir)->owner = NULL;
-	(*repo_dir)->description = NULL;
-	(*repo_dir)->url = NULL;
-	(*repo_dir)->path = NULL;
-
-	return NULL;
-}
-
 static const struct got_error *
 gotweb_get_repo_description(char **description, struct server *srv,
     const char *dirpath, int dir)
