@@ -88,7 +88,10 @@ struct got_reference {
 	struct got_lockfile *lf;
 	time_t mtime;
 
-	/* Cached timestamp for got_ref_cmp_by_commit_timestamp_descending() */
+	/*
+	 * Cached timestamp for got_ref_cmp_by_commit_timestamp_descending()
+	 * and got_ref_cmp_tags().
+	 */
 	time_t committer_time;
 };
 
@@ -649,70 +652,6 @@ got_ref_cmp_by_name(void *arg, int *cmp, struct got_reference *re1,
 	return NULL;
 }
 
-const struct got_error *
-got_ref_cmp_tags(void *arg, int *cmp, struct got_reference *ref1,
-    struct got_reference *ref2)
-{
-	const struct got_error *err = NULL;
-	struct got_repository *repo = arg;
-	struct got_object_id *id1, *id2 = NULL;
-	struct got_tag_object *tag1 = NULL, *tag2 = NULL;
-	struct got_commit_object *commit1 = NULL, *commit2 = NULL;
-	time_t time1, time2;
-
-	*cmp = 0;
-
-	err = got_ref_resolve(&id1, repo, ref1);
-	if (err)
-		return err;
-	err = got_object_open_as_tag(&tag1, repo, id1);
-	if (err) {
-		if (err->code != GOT_ERR_OBJ_TYPE)
-			goto done;
-		/* "lightweight" tag */
-		err = got_object_open_as_commit(&commit1, repo, id1);
-		if (err)
-			goto done;
-		time1 = got_object_commit_get_committer_time(commit1);
-	} else
-		time1 = got_object_tag_get_tagger_time(tag1);
-
-	err = got_ref_resolve(&id2, repo, ref2);
-	if (err)
-		goto done;
-	err = got_object_open_as_tag(&tag2, repo, id2);
-	if (err) {
-		if (err->code != GOT_ERR_OBJ_TYPE)
-			goto done;
-		/* "lightweight" tag */
-		err = got_object_open_as_commit(&commit2, repo, id2);
-		if (err)
-			goto done;
-		time2 = got_object_commit_get_committer_time(commit2);
-	} else
-		time2 = got_object_tag_get_tagger_time(tag2);
-
-	/* Put latest tags first. */
-	if (time1 < time2)
-		*cmp = 1;
-	else if (time1 > time2)
-		*cmp = -1;
-	else
-		err = got_ref_cmp_by_name(NULL, cmp, ref2, ref1);
-done:
-	free(id1);
-	free(id2);
-	if (tag1)
-		got_object_tag_close(tag1);
-	if (tag2)
-		got_object_tag_close(tag2);
-	if (commit1)
-		got_object_commit_close(commit1);
-	if (commit2)
-		got_object_commit_close(commit2);
-	return err;
-}
-
 static const struct got_error *
 get_committer_time(struct got_reference *ref, struct got_repository *repo)
 {
@@ -758,6 +697,37 @@ done:
 	return err;
 }
 
+const struct got_error *
+got_ref_cmp_tags(void *arg, int *cmp, struct got_reference *ref1,
+    struct got_reference *ref2)
+{
+	const struct got_error *err = NULL;
+	struct got_repository *repo = arg;
+
+	*cmp = 0;
+
+	if (ref1->committer_time == 0) {
+		err = get_committer_time(ref1, repo);
+		if (err)
+			return err;
+	}
+	if (ref2->committer_time == 0) {
+		err = get_committer_time(ref2, repo);
+		if (err)
+			return err;
+	}
+
+	/* Put latest tags first. */
+	if (ref1->committer_time < ref2->committer_time)
+		*cmp = 1;
+	else if (ref1->committer_time > ref2->committer_time)
+		*cmp = -1;
+	else
+		err = got_ref_cmp_by_name(NULL, cmp, ref2, ref1);
+
+	return err;
+}
+
 const struct got_error *
 got_ref_cmp_by_commit_timestamp_descending(void *arg, int *cmp,
     struct got_reference *ref1, struct got_reference *ref2)
@@ -886,6 +856,8 @@ gather_on_disk_refs(struct got_reflist_head *refs, const char *path_refs,
 	const struct got_error *err = NULL;
 	DIR *d = NULL;
 	char *path_subdir;
+	struct got_reference *ref;
+	struct got_reflist_entry *new;
 
 	while (subdir[0] == '/')
 		subdir++;
@@ -894,12 +866,40 @@ gather_on_disk_refs(struct got_reflist_head *refs, const char *path_refs,
 		return got_error_from_errno("asprintf");
 
 	d = opendir(path_subdir);
-	if (d == NULL)
-		goto done;
+	if (d == NULL) {
+		char *refname;
+
+		if (errno != ENOTDIR)
+			goto done;
+
+		/* This could be a regular on-disk reference file. */
+		free(path_subdir);
+		err = got_path_dirname(&path_subdir, subdir);
+		if (err)
+			return err;
+		err = got_path_basename(&refname, subdir);
+		if (err) {
+			free(path_subdir);
+			return err;
+		}
+		err = open_ref(&ref, path_refs, path_subdir, refname,
+		    0, got_repo_get_object_format(repo));
+		free(path_subdir);
+		free(refname);
+		if (err) {
+			if (err->code == GOT_ERR_NOT_REF)
+				return NULL;
+			return err;
+		}
+		err = got_reflist_insert(&new, refs, ref,
+		    cmp_cb, cmp_arg);
+		if (err || new == NULL /* duplicate */)
+			got_ref_close(ref);
+		return err;
+	}
 
 	for (;;) {
 		struct dirent *dent;
-		struct got_reference *ref;
 		char *child;
 		int type;
 
@@ -924,7 +924,6 @@ gather_on_disk_refs(struct got_reflist_head *refs, const char *path_refs,
 			if (err)
 				goto done;
 			if (ref) {
-				struct got_reflist_entry *new;
 				err = got_reflist_insert(&new, refs, ref,
 				    cmp_cb, cmp_arg);
 				if (err || new == NULL /* duplicate */)
@@ -954,6 +953,33 @@ done:
 	return err;
 }
 
+static int
+match_packed_ref(struct got_reference *ref, const char *ref_namespace)
+{
+	const char *name = got_ref_get_name(ref);
+	int namespace_is_absolute = (strncmp(ref_namespace, "refs/", 5) == 0);
+
+	if (namespace_is_absolute) {
+		return (strcmp(name, ref_namespace) == 0 ||
+		    got_path_is_child(name, ref_namespace,
+		    strlen(ref_namespace)));
+	}
+
+	/* Match all "subdirectories" as we do with on-disk refs. */
+	while (*name != '\0') {
+		while (*name == '/')
+			name++;
+		if (strcmp(name, ref_namespace) == 0 ||
+		    got_path_is_child(name, ref_namespace,
+		    strlen(ref_namespace)))
+			return 1;
+		while (*name != '\0' && *name != '/')
+			name++;
+	}
+
+	return 0;
+}
+
 const struct got_error *
 got_ref_list(struct got_reflist_head *refs, struct got_repository *repo,
     const char *ref_namespace, got_ref_cmp_cb cmp_cb, void *cmp_arg)
@@ -984,14 +1010,7 @@ got_ref_list(struct got_reflist_head *refs, struct got_repository *repo,
 			goto done;
 	} else {
 		/* Try listing a single reference. */
-		const char *refname = ref_namespace;
-		path_refs = get_refs_dir_path(repo, refname);
-		if (path_refs == NULL) {
-			err = got_error_from_errno("get_refs_dir_path");
-			goto done;
-		}
-		err = open_ref(&ref, path_refs, "", refname, 0,
-		    got_repo_get_object_format(repo));
+		err = got_ref_open(&ref, repo, ref_namespace, 0);
 		if (err) {
 			if (err->code != GOT_ERR_NOT_REF)
 				goto done;
@@ -1078,15 +1097,10 @@ got_ref_list(struct got_reflist_head *refs, struct got_repository *repo,
 			if (err)
 				goto done;
 			if (ref) {
-				if (ref_namespace) {
-					const char *name;
-					name = got_ref_get_name(ref);
-					if (!got_path_is_child(name,
-					    ref_namespace,
-					    strlen(ref_namespace))) {
-						got_ref_close(ref);
-						continue;
-					}
+				if (ref_namespace &&
+				    !match_packed_ref(ref, ref_namespace)) {
+					got_ref_close(ref);
+					continue;
 				}
 				err = got_reflist_insert(&new, refs, ref,
 				    cmp_cb, cmp_arg);
