@@ -24,6 +24,7 @@
 #include <stdint.h>
 #include <errno.h>
 #include <limits.h>
+#include <poll.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -135,7 +136,7 @@ recv_packfd(int *packfd, struct imsgbuf *ibuf)
 		goto done;
 	}
 
-	*packfd = imsg.fd;
+	*packfd = imsg_get_fd(&imsg);
 done:
 	imsg_free(&imsg);
 	return err;
@@ -212,7 +213,6 @@ send_their_ref(struct imsgbuf *ibuf, struct got_object_id *refid,
 	if (imsg_add(wbuf, refname, reflen) == -1)
 		return got_error_from_errno("imsg_add SEND_REMOTE_REF");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 	return got_privsep_flush_imsg(ibuf);
 }
@@ -296,7 +296,6 @@ send_ref_status(struct imsgbuf *ibuf, const char *refname, int success,
 	if (imsg_add(wbuf, errmsg, errmsglen) == -1)
 		return got_error_from_errno("imsg_add SEND_REF_STATUS");
 
-	wbuf->fd = -1;
 	imsg_close(ibuf, wbuf);
 	return got_privsep_flush_imsg(ibuf);
 }
@@ -362,7 +361,8 @@ send_pack(int fd, struct got_pathlist_head *refs,
 		return got_error(GOT_ERR_SEND_EMPTY);
 
 	while (1) {
-		err = got_pkt_readpkt(&n, fd, buf, sizeof(buf), chattygot);
+		err = got_pkt_readpkt(&n, fd, buf, sizeof(buf), chattygot,
+		    INFTIM);
 		if (err)
 			goto done;
 		if (n == 0)
@@ -549,7 +549,7 @@ send_pack(int fd, struct got_pathlist_head *refs,
 			goto done;
 	}
 
-	err = got_pkt_readpkt(&n, fd, buf, sizeof(buf), chattygot);
+	err = got_pkt_readpkt(&n, fd, buf, sizeof(buf), chattygot, INFTIM);
 	if (err)
 		goto done;
 	if (n >= 4 && strncmp(buf, "ERR ", 4) == 0) {
@@ -562,7 +562,8 @@ send_pack(int fd, struct got_pathlist_head *refs,
 	}
 
 	while (nsent > 0) {
-		err = got_pkt_readpkt(&n, fd, buf, sizeof(buf), chattygot);
+		err = got_pkt_readpkt(&n, fd, buf, sizeof(buf), chattygot,
+		    INFTIM);
 		if (err)
 			goto done;
 		if (n < 3) {
@@ -659,7 +660,7 @@ main(int argc, char **argv)
 		goto done;
 	}
 	memcpy(&send_req, imsg.data, sizeof(send_req));
-	sendfd = imsg.fd;
+	sendfd = imsg_get_fd(&imsg);
 	imsg_free(&imsg);
 
 	if (send_req.verbosity > 0)
